/*************************************************************************************
 *    M A R I T I M E  R E S E A R C H  I N S T I T U T E  N E T H E R L A N D S     *
 *************************************************************************************
 *    authors: Koos Huijssen, Christiaan M. Klaij                                    *
 *************************************************************************************
 *    content: Viewer for writing XML output                                         *
 *************************************************************************************/
#include <petsc-private/viewerimpl.h>
#include <petsc-private/fortranimpl.h>
#include "XMLViewer.h"
#include <stdbool.h>
#define VERBOSE 0

static int XMLSectionDepth            = 0;


/******************************************************************************************/
void PetscViewerInitASCII_XMLRoutine( PetscViewer *viewer, PetscErrorCode *ierr)
{
  PetscViewer v;
  PetscPatchDefaultViewers_Fortran(viewer,v);
  *ierr = PetscViewerInitASCII_XML(v);
}


/******************************************************************************************/
/* Initialize a viewer to XML, and initialize the XMLDepth static parameter
 */
#undef __FUNCT__
#define __FUNCT__ "PetscViewerInitASCII_XML"
PetscErrorCode PetscViewerInitASCII_XML(PetscViewer viewer)
{
  PetscErrorCode ierr;
  MPI_Comm       comm;
  FILE           *fd;

  PetscFunctionBegin;

  ierr = PetscViewerSetFormat(viewer, PETSC_VIEWER_ASCII_IMPL); CHKERRQ(ierr);
  // We cannot add another PetscViewerFormat beyond the existing enum list, hence we are using ASCII_IMPL
  // for our purpose.
  
  ierr = PetscObjectGetComm((PetscObject)viewer,&comm);CHKERRQ(ierr);
  ierr = PetscViewerASCIIGetPointer(viewer,&fd);CHKERRQ(ierr);
  ierr = PetscFPrintf(comm, fd, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");CHKERRQ(ierr);
  ierr = PetscFPrintf(comm, fd, "<?xml-stylesheet type=\"text/xsl\" href=\"performance_xml2html.xsl\"?>\n");CHKERRQ(ierr);

  XMLSectionDepth = 0;
  
  ierr = PetscXMLStartSection(comm, fd, "root", "");CHKERRQ(ierr);

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* Initialize a viewer to XML, and initialize the XMLDepth static parameter
 */
#undef __FUNCT__
#define __FUNCT__ "PetscViewerFinalASCII_XML"
PetscErrorCode PetscViewerFinalASCII_XML(PetscViewer *viewer)
{
  PetscErrorCode ierr;
  MPI_Comm       comm;
  FILE           *fd;

  PetscFunctionBegin;

  ierr = PetscObjectGetComm((PetscObject)*viewer,&comm);CHKERRQ(ierr);
  ierr = PetscViewerASCIIGetPointer(*viewer,&fd);CHKERRQ(ierr);
  ierr = PetscXMLEndSection(comm, fd, "root");CHKERRQ(ierr);

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* Check whether the viewer is set to XML format
 */
#undef __FUNCT__
#define __FUNCT__ "PetscViewerCheckASCII_XMLFormat"
PetscErrorCode PetscViewerCheckASCII_XMLFormat(PetscViewer viewer, bool *XMLFormat)
{
  PetscErrorCode ierr;
  PetscViewerFormat format;
  PetscFunctionBegin;

  ierr = PetscViewerGetFormat(viewer, &format);CHKERRQ(ierr);

  *XMLFormat = (format == PETSC_VIEWER_ASCII_IMPL); 

  PetscFunctionReturn(0);
}


/******************************************************************************************/
void PetscViewerXMLStartSectionRoutine(PetscViewer *viewer, const char *name, const char *desc, PetscErrorCode *ierr)
{
  PetscViewer v;

  PetscPatchDefaultViewers_Fortran(viewer,v);
  *ierr = PetscViewerXMLStartSection(v, name, desc);
}
/******************************************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscViewerXMLStartSection"
PetscErrorCode PetscViewerXMLStartSection(PetscViewer viewer, const char *name, const char *desc)
{
  PetscErrorCode ierr;
  MPI_Comm       comm;
  FILE           *fd;

  PetscFunctionBegin;

  ierr = PetscObjectGetComm((PetscObject)viewer,&comm);CHKERRQ(ierr);
  ierr = PetscViewerASCIIGetPointer(viewer,&fd);CHKERRQ(ierr);
  ierr = PetscXMLStartSection(comm, fd, name, desc);

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* Start an XML Section in the output file; increment XMLSectionDepth by one.
 */
#undef __FUNCT__
#define __FUNCT__ "PetscXMLStartSection"
PetscErrorCode PetscXMLStartSection(MPI_Comm comm, FILE *fd, const char *name, const char *desc)
{
  PetscErrorCode ierr;

  PetscFunctionBegin;

  if(!strcmp(desc,"")) {
    ierr = PetscFPrintf(comm, fd, "%*s<%s>\n", XMLSectionDepth, "", name);CHKERRQ(ierr);
  } else {
    ierr = PetscFPrintf(comm, fd, "%*s<%s desc=\"%s\">\n", XMLSectionDepth, "", name, desc);CHKERRQ(ierr);
  }
  XMLSectionDepth += 2;

  PetscFunctionReturn(0);
}


/******************************************************************************************/
void PetscViewerXMLEndSectionRoutine(PetscViewer *viewer, const char *name, PetscErrorCode *ierr)
{
  PetscViewer v;

  PetscPatchDefaultViewers_Fortran(viewer,v);
  *ierr = PetscViewerXMLEndSection(v, name);
}
/******************************************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscViewerXMLEndSection"
PetscErrorCode PetscViewerXMLEndSection(PetscViewer viewer, const char *name)
{
  PetscErrorCode ierr;
  MPI_Comm       comm;
  FILE           *fd;

  PetscFunctionBegin;

  ierr = PetscObjectGetComm((PetscObject)viewer,&comm);CHKERRQ(ierr);
  ierr = PetscViewerASCIIGetPointer(viewer,&fd);CHKERRQ(ierr);
  ierr = PetscXMLEndSection(comm, fd, name);

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* End an XML Section in the output file; decrement XMLSectionDepth by one.
 */
#undef __FUNCT__
#define __FUNCT__ "PetscXMLEndSection"
PetscErrorCode PetscXMLEndSection(MPI_Comm comm, FILE *fd, const char *name)
{
  PetscErrorCode ierr;

  PetscFunctionBegin;

  XMLSectionDepth -= 2;
  if (XMLSectionDepth<0) XMLSectionDepth = 0;
  ierr = PetscFPrintf(comm, fd, "%*s</%s>\n", XMLSectionDepth, "", name);CHKERRQ(ierr);

  PetscFunctionReturn(0);
}


/******************************************************************************************/
void PetscViewerXMLPutStringRoutine(PetscViewer *viewer, const char *name, const char *desc, const char *value, PetscErrorCode *ierr)
{
  PetscViewer v;

  PetscPatchDefaultViewers_Fortran(viewer,v);
  *ierr = PetscViewerXMLPutString(v, name, desc, value);
}
/******************************************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscViewerXMLPutString"
PetscErrorCode PetscViewerXMLPutString(PetscViewer viewer, const char *name, const char *desc, const char *value)
{
  PetscErrorCode ierr;
  MPI_Comm       comm;
  FILE           *fd;

  PetscFunctionBegin;

  ierr = PetscObjectGetComm((PetscObject)viewer,&comm);CHKERRQ(ierr);
  ierr = PetscViewerASCIIGetPointer(viewer,&fd);CHKERRQ(ierr);
  ierr = PetscXMLPutString(comm, fd, name, desc, value);

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* Print an XML Element with a string in the output file.
 */
#undef __FUNCT__
#define __FUNCT__ "PetscXMLPutString"
PetscErrorCode PetscXMLPutString(MPI_Comm comm, FILE *fd, const char *name, const char *desc, const char * value)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;

  if(!strcmp(desc,"")) {
    ierr = PetscFPrintf(comm, fd, "%*s<%s>%s</%s>\n", XMLSectionDepth, "", name, value, name);CHKERRQ(ierr);
  } else {
    ierr = PetscFPrintf(comm, fd, "%*s<%s desc=\"%s\">%s</%s>\n", XMLSectionDepth, "", name, desc, value, name);CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}


/******************************************************************************************/
void PetscViewerXMLPutIntRoutine(PetscViewer *viewer, const char *name, const char *desc, const int *value, PetscErrorCode *ierr)
{
  PetscViewer v;

  PetscPatchDefaultViewers_Fortran(viewer,v);
  *ierr = PetscViewerXMLPutInt(v, name, desc, *value);
}
/******************************************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscViewerXMLPutInt"
PetscErrorCode PetscViewerXMLPutInt(PetscViewer viewer, const char *name, const char *desc, const int value)
{
  PetscErrorCode ierr;
  MPI_Comm       comm;
  FILE           *fd;

  PetscFunctionBegin;

  ierr = PetscObjectGetComm((PetscObject)viewer,&comm);CHKERRQ(ierr);
  ierr = PetscViewerASCIIGetPointer(viewer,&fd);CHKERRQ(ierr);
  ierr = PetscXMLPutInt(comm, fd, name, desc, value);

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* Print an XML Element with an integer value in the output file.
 */
#undef __FUNCT__
#define __FUNCT__ "PetscXMLPutInt"
PetscErrorCode PetscXMLPutInt(MPI_Comm comm, FILE *fd, const char *name, const char *desc, const int value)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;

  if(!strcmp(desc,"")) {
    ierr = PetscFPrintf(comm, fd, "%*s<%s>%d</%s>\n", XMLSectionDepth, "", name, value, name);CHKERRQ(ierr);
  } else {
    ierr = PetscFPrintf(comm, fd, "%*s<%s desc=\"%s\">%d</%s>\n", XMLSectionDepth, "", name, desc, value, name);CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}


/******************************************************************************************/
void PetscViewerXMLPutDoubleRoutine(PetscViewer *viewer, const char *name, const char *desc, const PetscLogDouble *value, const char *format, PetscErrorCode *ierr)
{
  PetscViewer v;

  PetscPatchDefaultViewers_Fortran(viewer,v);
  *ierr = PetscViewerXMLPutDouble(v, name, desc, *value, format);
}
/******************************************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscViewerXMLPutDouble"
PetscErrorCode PetscViewerXMLPutDouble(PetscViewer viewer, const char *name, const char *desc, const PetscLogDouble value, const char *format)
{
  PetscErrorCode ierr;
  MPI_Comm       comm;
  FILE           *fd;

  PetscFunctionBegin;

  ierr = PetscObjectGetComm((PetscObject)viewer,&comm);CHKERRQ(ierr);
  ierr = PetscViewerASCIIGetPointer(viewer,&fd);CHKERRQ(ierr);
  ierr = PetscXMLPutDouble(comm, fd, name, desc, value, format);

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* Print an XML Element with a double value in the output file.
 */
#undef __FUNCT__
#define __FUNCT__ "PetscXMLPutDouble"
PetscErrorCode PetscXMLPutDouble(MPI_Comm comm, FILE *fd, const char *name, const char *desc, const PetscLogDouble value, const char *format)
{
  PetscErrorCode ierr;
  char buffer[128];
  PetscFunctionBegin;

  if(!strcmp(desc,"")) {
    sprintf(buffer, "%*s<%s>%s</%s>\n", XMLSectionDepth, "", name, format, name);
  } else {
    sprintf(buffer, "%*s<%s desc=\"%s\">%s</%s>\n", XMLSectionDepth, "", name, desc, format, name);CHKERRQ(ierr);
  }
  ierr = PetscFPrintf(comm, fd, buffer, value);CHKERRQ(ierr);

  PetscFunctionReturn(0);
}  
