/*************************************************************************************
 *    M A R I T I M E  R E S E A R C H  I N S T I T U T E  N E T H E R L A N D S     *
 *************************************************************************************
 *    authors: Bas van 't Hof, Koos Huijssen, Christiaan M. Klaij                    *
 *************************************************************************************
 *    content: Support for nested PetscTimers                                        *
 *************************************************************************************/
#include <petsclog.h> 
#include <petsc-private/logimpl.h>
#include <petsctime.h>
#include <petscviewer.h>
#include <petscthreadcomm.h>
#include "myLogEvent.h" 
#include "XMLViewer.h"
#include <petsc-private/fortranimpl.h>
#include <stdbool.h>

#define VERBOSE 0

/*************************************************************************************
 * Support for nested PetscTimers
 *
 * PetscTimers keep track of a lot of useful information: Wall clock times, 
 * message passing statistics, flop counts.  Information about the nested structure 
 * of the timers is lost. Example: 
 *
 * 7:30   Start: awake
 * 7:30      Start: morning routine
 * 7:40         Start: eat
 * 7:49         Done:  eat
 * 7:43      Done:  morning routine
 * 8:15      Start: work
 * 12:15        Start: eat
 * 12:45        Done:  eat
 * 16:00     Done:  work
 * 16:30     Start: evening routine
 * 18:30        Start: eat
 * 19:15        Done:  eat
 * 22:00     Done:  evening routine
 * 22:00  Done:  awake
 *
 * Petsc timers provide the following timer results:
 *
 *    awake:              1 call    14:30 hours
 *    morning routine:    1 call     0:13 hours
 *    eat:                3 calls    1:24 hours
 *    work:               1 call     7:45 hours
 *    evening routine     1 call     5:30 hours
 *
 * Nested timers can be used to get the following table:
 *
 *   [1 call]: awake                14:30 hours
 *   [1 call]:    morning routine         0:13 hours         ( 2 % of awake)
 *   [1 call]:       eat                       0:09 hours         (69 % of morning routine)
 *                   rest (morning routine)    0:04 hours         (31 % of morning routine)
 *   [1 call]:    work                    7:45 hours         (53 % of awake)
 *   [1 call]:       eat                       0:30 hours         ( 6 % of work)
 *                   rest (work)               7:15 hours         (94 % of work)
 *   [1 call]:    evening routine         5:30 hours         (38 % of awake)
 *   [1 call]:       eat                       0:45 hours         (14 % of evening routine)
 *                   rest (evening routine)    4:45 hours         (86 % of morning routine)
 *
 * We ignore the concept of 'stages', because these seem to be conflicting notions, or at least, 
 * the nested timers make the stages unnecessary.
 *
 */

/*****************************************************************************************************
 * Data structures for keeping track of nested timers:
 *
 *   nestedEvents: information about the timers that have actually been activated
 *   dftParentActive: if a timer is started now, it is part of (nested inside) the dftParentActive
 *
 * The Default-timers are used to time the nested timers. Every nested timer corresponds to 
 * (one or more) default timers, where one of the default timers has the same event-id as the 
 * nested one.
 *
 * Because of the risk of confusion between nested timer ids and default timer ids, we 
 * introduce a typedef for nested events (NestedEventId) and use the existing type PetscLogEvent 
 * only for default events. Also, all nested event variables are prepended with 'nst', and
 * default timers with 'dft'.
 */
typedef PetscLogEvent NestedEventId; 
typedef struct {
  NestedEventId   nstEvent;         /* event-code for this nested event, argument 'event' in PetscLogEventStartNested */
                                  
  size_t nParents;                  /* number of 'dftParents': the default timer which was the dftParentActive when this nested 
                                       timer was activated */
  PetscLogEvent  *dftParentsSorted; /* The default timers which were the dftParentActive when this nested event was started */
  PetscLogEvent  *dftEvents;        /* The default timers which represent the different 'instances' of this nested event */

  PetscLogEvent  *dftParents;       /* The default timers which were the dftParentActive when this nested event was started */
  PetscLogEvent  *dftEventsSorted;  /* The default timers which represent the different 'instances' of this nested event */
  
} PetscNestedEvent;

static PetscLogEvent dftParentActive  = 0;
static size_t nNestedEvents           = 0;
static size_t nNestedEventsAllocated  = 0;
static PetscNestedEvent *nestedEvents = NULL;
static PetscLogDouble threshTime      = 0.1;
static bool useXMLFormat              = false;

/*****************************************************************************************************/
static PetscErrorCode PetscLogEventBeginNested(NestedEventId nstEvent, int t, PetscObject o1, PetscObject o2, PetscObject o3, PetscObject o4);
static PetscErrorCode PetscLogEventEndNested(NestedEventId nstEvent, int t, PetscObject o1, PetscObject o2, PetscObject o3, PetscObject o4);
/*****************************************************************************************************
 * Initialize the data structures for the nested timers */
#undef __FUNCT__
#define __FUNCT__ "PetscLogInitializeNested"
PetscErrorCode PetscLogInitializeNested()
{
# if VERBOSE
     PetscMPIInt        rank;
# endif
  PetscErrorCode    ierr;
  PetscFunctionBegin;
# if VERBOSE
     ierr = MPI_Comm_rank(PETSC_COMM_WORLD, &rank);CHKERRQ(ierr);
     printf("[%d] Welcome to %s\n", rank, __FUNCT__);
# endif
  if (nestedEvents) SETERRQ(PETSC_COMM_SELF,PETSC_ERR_COR,"nestedEvents already allocated");

  nNestedEventsAllocated=10;
  ierr = PetscMalloc(nNestedEventsAllocated*sizeof(PetscNestedEvent),&nestedEvents);CHKERRQ(ierr);


  dftParentActive = 0;
  nNestedEvents =1;

  /* 'Awake' is nested event 0. It has no parents */
  nestedEvents[0].nstEvent          = 0;
  nestedEvents[0].nParents          = 0;
  nestedEvents[0].dftParentsSorted  = NULL;
  nestedEvents[0].dftEvents         = NULL;
  nestedEvents[0].dftParents        = NULL;
  nestedEvents[0].dftEventsSorted   = NULL;

  ierr = PetscLogSet(PetscLogEventBeginNested, PetscLogEventEndNested); CHKERRQ(ierr); 

  PetscFunctionReturn(0);
}

/*****************************************************************************************************
 * Delete the data structures for the nested timers */
#undef __FUNCT__
#define __FUNCT__ "PetscLogFreeNested"
PetscErrorCode PetscLogFreeNested()
{
  PetscErrorCode    ierr;
  size_t            i;
  PetscFunctionBegin;

  if (!nestedEvents) SETERRQ(PETSC_COMM_SELF,PETSC_ERR_COR,"nestedEvents does not exist");

  for (i=0; i<nNestedEvents; i++) {
      ierr = PetscFree(nestedEvents[i].dftParentsSorted);CHKERRQ(ierr);
      ierr = PetscFree(nestedEvents[i].dftEventsSorted);CHKERRQ(ierr);
      ierr = PetscFree(nestedEvents[i].dftParents);CHKERRQ(ierr);
      ierr = PetscFree(nestedEvents[i].dftEvents);CHKERRQ(ierr);
  }
  ierr = PetscFree(nestedEvents);CHKERRQ(ierr);
  nestedEvents = NULL;
  nNestedEvents = 0;
  nNestedEventsAllocated=0;
  //ierr = PetscMallocDump(stdout);
  PetscFunctionReturn(0);
}


/*****************************************************************************************************
 * UTILITIES: FIND STUFF IN SORTED ARRAYS
 *
 * Utility: find a default timer in a sorted array */
#undef __FUNCT__
#define __FUNCT__ "PetscLogEventFindDefaultTimer"
static PetscErrorCode PetscLogEventFindDefaultTimer(
            const PetscLogEvent dftIndex,    /* index to be found */
            const PetscLogEvent *dftArray,  /* sorted array of PetscLogEvent-ids */
            const size_t narray,   /* dimension of dftArray */
            size_t *entry          /* entry in the array where dftIndex may be found; 
                                    *   if dftArray[entry] != dftIndex, then dftIndex is not part of dftArray  
                                    *   In that case, the dftIndex can be inserted at this entry.
                                    */
) {
  PetscFunctionBegin;

  if (narray==0 || dftIndex <= dftArray[0]) { 
      *entry = 0;
  } else if (dftIndex > dftArray[narray-1]) {
      *entry = narray;
  } else {
     size_t ihigh=narray-1,  ilow=0;
     while (ihigh>ilow) {
        const size_t imiddle = (ihigh+ilow)/2;
        if (dftArray[imiddle] > dftIndex) {
           ihigh=imiddle;
        } else if (dftArray[imiddle]<dftIndex) {
           ilow =imiddle+1;
        } else {
           ihigh=imiddle;
           ilow =imiddle;
        }
     }
     *entry = ihigh;
  }

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* Utility: find the nested event with given identification */
#undef __FUNCT__
#define __FUNCT__ "PetscLogEventFindNestedTimer"
static PetscErrorCode PetscLogEventFindNestedTimer(
            const NestedEventId nstEvent, /* Nested event to be found */
            size_t *entry                 /* entry in the nestedEvents where nstEvent may be found;
                                             if nestedEvents[entry].nstEvent != nstEvent, then index is not part of iarray  */
) {
  PetscFunctionBegin;

  if (nNestedEvents==0 || nstEvent <= nestedEvents[0].nstEvent) { 
      *entry = 0;
  } else if (nstEvent > nestedEvents[nNestedEvents-1].nstEvent) {
      *entry = nNestedEvents;
  } else {
      size_t ihigh=nNestedEvents-1,  ilow=0;
      while (ihigh>ilow) {
         const size_t imiddle = (ihigh+ilow)/2;
         if (nestedEvents[imiddle].nstEvent > nstEvent) {
            ihigh=imiddle;
         } else if (nestedEvents[imiddle].nstEvent<nstEvent) {
            ilow =imiddle+1;
         } else {
            ihigh=imiddle;
            ilow =imiddle;
         }
      }
      *entry = ihigh;
  }
  PetscFunctionReturn(0);
}

/******************************************************************************************/
/* Start a nested event */
#undef __FUNCT__
#define __FUNCT__ "PetscLogEventBeginNested"
static PetscErrorCode PetscLogEventBeginNested(NestedEventId nstEvent, int t, PetscObject o1, PetscObject o2, PetscObject o3, PetscObject o4)
{
  PetscErrorCode    ierr;
# if VERBOSE 
     PetscStageLog      stageLog;
     const char         *name;
     PetscMPIInt        rank;
# endif

  PetscFunctionBegin;

# if VERBOSE 
     ierr = MPI_Comm_rank(PETSC_COMM_WORLD, &rank);CHKERRQ(ierr);
     ierr      = PetscLogGetStageLog(&stageLog);CHKERRQ(ierr);
     name      = stageLog->eventLog->eventInfo[(PetscLogEvent) nstEvent].name;
# endif

# if VERBOSE 
     printf("[%d] Welcome to %s\n", rank, __FUNCT__);
     printf("[%d]   starting nested event %d (%s) from %d\n", 
             rank, nstEvent, name, dftParentActive);
# endif

  size_t entry, pentry, tentry;
  ierr = PetscLogEventFindNestedTimer(nstEvent, &entry); CHKERRQ(ierr); 

  PetscLogEvent dftEvent;

  if (entry>=nNestedEvents || nestedEvents[entry].nstEvent != nstEvent) {
  /* Nested event doesn't exist yet: create it */
     size_t i;

     if (nNestedEvents==nNestedEventsAllocated) {
     /* Enlarge and re-allocate nestedEvents if needed */
        PetscNestedEvent *tmp = nestedEvents;
        ierr = PetscMalloc(2*nNestedEvents*sizeof(PetscNestedEvent),&nestedEvents);CHKERRQ(ierr);
        nNestedEventsAllocated*=2;
        memcpy(nestedEvents, tmp, nNestedEvents*sizeof(PetscNestedEvent)); 
        ierr = PetscFree(tmp);CHKERRQ(ierr);
     }
     
     /* Clear space in nestedEvents for new nested event */
     nNestedEvents++;
     for (i = nNestedEvents-1; i>entry; i--) {
         nestedEvents[i] = nestedEvents[i-1];
     }

     /* Create event in nestedEvents */
     nestedEvents[entry].nstEvent = nstEvent;
     nestedEvents[entry].nParents=1;
     ierr = PetscMalloc(sizeof(PetscLogEvent),&nestedEvents[entry].dftParentsSorted); CHKERRQ(ierr);
     ierr = PetscMalloc(sizeof(PetscLogEvent),&nestedEvents[entry].dftEvents);        CHKERRQ(ierr);
     ierr = PetscMalloc(sizeof(PetscLogEvent),&nestedEvents[entry].dftParents);       CHKERRQ(ierr);
     ierr = PetscMalloc(sizeof(PetscLogEvent),&nestedEvents[entry].dftEventsSorted);  CHKERRQ(ierr);

     /* Fill in new event */
     pentry = 0;
     dftEvent = (PetscLogEvent) nstEvent;

     nestedEvents[entry].nstEvent                 = nstEvent;
     nestedEvents[entry].dftParents[pentry]       = dftParentActive;
     nestedEvents[entry].dftEvents[pentry]        = dftEvent;
     nestedEvents[entry].dftParentsSorted[pentry] = dftParentActive;
     nestedEvents[entry].dftEventsSorted[pentry]  = dftEvent;

#    if VERBOSE
         printf("[%d]   nested event %d corresponds to its own default timer: %d\n", 
                rank, nstEvent, dftEvent);
#    endif

  } else {
  /* Nested event exists: find current dftParentActive among parents */
     PetscLogEvent *dftParentsSorted = nestedEvents[entry].dftParentsSorted; 
     PetscLogEvent *dftEvents        = nestedEvents[entry].dftEvents; 
     size_t nParents = nestedEvents[entry].nParents;

     ierr = PetscLogEventFindDefaultTimer( dftParentActive, dftParentsSorted, nParents, &pentry); CHKERRQ(ierr); 

     if (pentry>=nParents || dftParentActive != dftParentsSorted[pentry]) {
     /* dftParentActive not in the list: add it to the list */
        size_t i;
        PetscLogEvent *dftParents      = nestedEvents[entry].dftParents; 
        PetscLogEvent *dftEventsSorted = nestedEvents[entry].dftEventsSorted; 

        /* Register a new default timer */
        char name[100]; 
        sprintf(name, "%d -> %d", (int) dftParentActive, (int) nstEvent);
        ierr = PetscLogEventRegister(name, 0, &dftEvent);
        ierr = PetscLogEventFindDefaultTimer( dftEvent, dftEventsSorted, nParents, &tentry); CHKERRQ(ierr); 

        /* Reallocate parents and dftEvents to make space for new parent */
        ierr = PetscMalloc((1+nParents)*sizeof(PetscLogEvent),&nestedEvents[entry].dftParentsSorted); CHKERRQ(ierr);
        ierr = PetscMalloc((1+nParents)*sizeof(PetscLogEvent),&nestedEvents[entry].dftEventsSorted);  CHKERRQ(ierr);
        ierr = PetscMalloc((1+nParents)*sizeof(PetscLogEvent),&nestedEvents[entry].dftParents);       CHKERRQ(ierr);
        ierr = PetscMalloc((1+nParents)*sizeof(PetscLogEvent),&nestedEvents[entry].dftEvents);        CHKERRQ(ierr);
        memcpy(nestedEvents[entry].dftParentsSorted, dftParentsSorted, nParents*sizeof(PetscLogEvent)); 
        memcpy(nestedEvents[entry].dftEventsSorted,  dftEventsSorted,  nParents*sizeof(PetscLogEvent)); 
        memcpy(nestedEvents[entry].dftParents,       dftParents,       nParents*sizeof(PetscLogEvent)); 
        memcpy(nestedEvents[entry].dftEvents,        dftEvents,        nParents*sizeof(PetscLogEvent)); 
        ierr = PetscFree(dftParents);CHKERRQ(ierr);
        ierr = PetscFree(dftEvents );CHKERRQ(ierr);
        ierr = PetscFree(dftParentsSorted);CHKERRQ(ierr);
        ierr = PetscFree(dftEventsSorted );CHKERRQ(ierr);

        dftParents       = nestedEvents[entry].dftParents;
        dftEvents        = nestedEvents[entry].dftEvents;
        dftParentsSorted = nestedEvents[entry].dftParentsSorted;
        dftEventsSorted  = nestedEvents[entry].dftEventsSorted;

        nestedEvents[entry].nParents++;
        nParents++;

        for (i = nParents-1; i>pentry; i--) {
            dftParentsSorted[i] = dftParentsSorted[i-1];
            dftEvents[i]        = dftEvents[i-1];
        }
        for (i = nParents-1; i>tentry; i--) {
            dftParents[i]      = dftParents[i-1];
            dftEventsSorted[i] = dftEventsSorted[i-1];
        }

        /* Fill in the new default timer */
        dftParentsSorted[pentry] = dftParentActive;
        dftEvents[pentry]        = dftEvent;
        dftParents[tentry]       = dftParentActive;
        dftEventsSorted[tentry]  = dftEvent;

#       if VERBOSE 
            if (rank <= 1) {
               printf("[%d]   nested event %d corresponds to (new) default timer %d (%s)\n", 
                     rank, nstEvent, dftEvent, name);
            }
#       endif
        
     } else {
     /* dftParentActive was found: find the corresponding default 'dftEvent'-timer */
        dftEvent = nestedEvents[entry].dftEvents[pentry];
#       if VERBOSE 
            if (rank <= 1) {
               printf("[%d]   nested event %d corresponds to (existing) default timer %d\n", 
                     rank, nstEvent, dftEvent);
            }
#       endif
     }
  }

  /* Start the default 'dftEvent'-timer and update the dftParentActive */
  ierr = PetscLogEventBeginDefault(dftEvent,t,o1,o2,o3,o4);    CHKERRQ(ierr); 
  dftParentActive = dftEvent;

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* End a nested event */
#undef __FUNCT__
#define __FUNCT__ "PetscLogEventEndNested"
static PetscErrorCode PetscLogEventEndNested(NestedEventId nstEvent, int t, PetscObject o1, PetscObject o2, PetscObject o3, PetscObject o4)
{
  PetscMPIInt     rank;
  PetscErrorCode  ierr;
  size_t          entry, pentry, nParents;
  PetscLogEvent  *dftEventsSorted;

  PetscFunctionBegin;

# if VERBOSE 
     ierr = MPI_Comm_rank(PETSC_COMM_WORLD, &rank);CHKERRQ(ierr);
     printf("[%d] Welcome to %s\n", rank, __FUNCT__);
     printf("[%d]   ending nested event %d\n", rank, nstEvent);
# endif

  /* Find the nested event */
  ierr = PetscLogEventFindNestedTimer(nstEvent, &entry); CHKERRQ(ierr); 
  if (entry>=nNestedEvents || nestedEvents[entry].nstEvent != nstEvent) {
      ierr = MPI_Comm_rank(PETSC_COMM_WORLD, &rank);CHKERRQ(ierr);
      printf("[%d]   ERROR: event %d not found\n", rank, nstEvent);
      SETERRQ(PETSC_COMM_SELF,PETSC_ERR_ARG_WRONGSTATE, "Logging event had unbalanced begin/end pairs");
  }

  dftEventsSorted = nestedEvents[entry].dftEventsSorted;
  nParents            = nestedEvents[entry].nParents;

  /* Find the current default timer among the 'dftEvents' of this event */
  ierr = PetscLogEventFindDefaultTimer( dftParentActive, dftEventsSorted, nParents, &pentry); CHKERRQ(ierr); 

  if (pentry>=nParents || dftEventsSorted[pentry] != dftParentActive) {
      int i;
      PetscEventRegInfo *eventInfo;
      PetscStageLog      stageLog;

      ierr = MPI_Comm_rank(PETSC_COMM_WORLD, &rank);CHKERRQ(ierr);
      ierr = PetscLogGetStageLog(&stageLog);CHKERRQ(ierr);
      eventInfo = stageLog->eventLog->eventInfo;

      printf("[%d] ERROR: active parent is %d, but we seem to be closing %d\n", rank, dftParentActive,
             dftEventsSorted[pentry]);

      for (i=0; i<nNestedEvents; i++) {
          int j;
          for (j=0; j<nestedEvents[i].nParents; j++) { 
              if (dftEventsSorted[j]==dftParentActive) {
                  printf("[%d] active parent is '%s', but we seem to be closing '%s'\n",
                         rank, eventInfo[dftEventsSorted[j]].name, eventInfo[(PetscLogEvent) nstEvent].name);
              }
          }
      }

      for (i=0; i<nestedEvents[entry].nParents; i++) {
         printf("[%d]  nested timer %d is default timer %d\n", rank, nstEvent, nestedEvents[entry].dftEventsSorted[i] );
      }

      SETERRQ(PETSC_COMM_SELF,PETSC_ERR_ARG_WRONGSTATE, "Logging event had unbalanced begin/end pairs");
  }

  /* Stop the default timer and update the dftParentActive */
  ierr = PetscLogEventEndDefault(dftParentActive,t,o1,o2,o3,o4);
  dftParentActive = nestedEvents[entry].dftParents[pentry]; 
# if VERBOSE
     printf("[%d]   back to dftParent %d\n", rank, dftParentActive);
# endif

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* Set the threshold time for logging the events 
 */
#undef __FUNCT__
#define __FUNCT__ "PetscLogSetThreshold"
PetscErrorCode PetscLogSetThreshold(const PetscLogDouble newThresh, PetscLogDouble *oldThresh) {
  PetscFunctionBegin;
  *oldThresh = threshTime;
  threshTime = newThresh;
  PetscFunctionReturn(0);
} 


/******************************************************************************************/
/* Print the specification of the executable: what is it called, who ran it, on which host, 
 * how many processes.
 */
#undef __FUNCT__
#define __FUNCT__ "PetscPrintExeSpecs"
static PetscErrorCode  PetscPrintExeSpecs(MPI_Comm comm,  FILE *fd)
{
  PetscErrorCode     ierr;
  char               arch[128],hostname[128],username[128],pname[PETSC_MAX_PATH_LEN],date[128];
  char               version[256], buildoptions[128];
  PetscInt           nthreads;
  PetscMPIInt        size;

  PetscFunctionBegin;
  ierr = PetscGetArchType(arch,sizeof(arch));CHKERRQ(ierr);
  ierr = PetscGetHostName(hostname,sizeof(hostname));CHKERRQ(ierr);
  ierr = PetscGetUserName(username,sizeof(username));CHKERRQ(ierr);
  ierr = PetscGetProgramName(pname,sizeof(pname));CHKERRQ(ierr);
  ierr = PetscGetDate(date,sizeof(date));CHKERRQ(ierr);
  ierr = PetscGetVersion(version,sizeof(version));CHKERRQ(ierr);
  ierr = MPI_Comm_size(comm, &size);CHKERRQ(ierr);
  ierr = PetscThreadCommGetNThreads(PETSC_COMM_WORLD,&nthreads);CHKERRQ(ierr);

  if(!useXMLFormat) {

    if (size == 1) {
      ierr = PetscFPrintf(comm,fd,"%s on a %s named %s with %d process, by %s %s\n", pname, arch, hostname, size, username, date);CHKERRQ(ierr);
    } else {
      ierr = PetscFPrintf(comm,fd,"%s on a %s named %s with %d processes, by %s %s\n", pname, arch, hostname, size, username, date);CHKERRQ(ierr);
    }
  
    if (nthreads > 1) {
      ierr = PetscFPrintf(comm,fd,"With %d threads per MPI_Comm\n", (int)nthreads);CHKERRQ(ierr);
    }
  
    ierr = PetscFPrintf(comm, fd, "Using %s\n", version);CHKERRQ(ierr);

  } else {

    ierr = PetscXMLStartSection(comm, fd, "runspecification", "Run Specification");CHKERRQ(ierr);
    ierr = PetscXMLPutString(   comm, fd, "executable"  , "Executable"   , pname );CHKERRQ(ierr);
    ierr = PetscXMLPutString(   comm, fd, "architecture", "Architecture" , arch );CHKERRQ(ierr);
    ierr = PetscXMLPutString(   comm, fd, "hostname"    , "Host"         , hostname);CHKERRQ(ierr);
    ierr = PetscXMLPutInt(      comm, fd, "nprocesses"  , "Number of processes", size );CHKERRQ(ierr);
    ierr = PetscXMLPutInt(      comm, fd, "nthreads"    , "Number of threads"   , nthreads);CHKERRQ(ierr);
    ierr = PetscXMLPutString(   comm, fd, "user"        , "Run by user"  , username);CHKERRQ(ierr);
    ierr = PetscXMLPutString(   comm, fd, "date"        , "Started at"   , date);CHKERRQ(ierr);
    ierr = PetscXMLPutString(   comm, fd, "petscrelease", "Petsc Release", version);CHKERRQ(ierr);
#if defined(PETSC_USE_DEBUG)
#  if defined(PETSC_USE_COMPLEX) && !defined(PETSC_USE_FORTRAN_KERNELS)
    sprintf(buildoptions, "Debug, ComplexC++Kernels");
#  else
    sprintf(buildoptions, "Debug");
#  endif
#else
#  if defined(PETSC_USE_COMPLEX) && !defined(PETSC_USE_FORTRAN_KERNELS)
    sprintf(buildoptions, "ComplexC++Kernels");
#  else
    sprintf(buildoptions, "");
#  endif
#endif
    if (strlen(buildoptions)>0) {
      ierr = PetscXMLPutString(comm, fd, "petscbuildoptions", "Petsc build options", buildoptions);CHKERRQ(ierr);
    }
    ierr = PetscXMLEndSection(comm, fd, "runspecification");CHKERRQ(ierr);

  }

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* Print the global performance: max, max/min, average and total of 
 *      time, objects, flops, flops/sec, memory, MPI messages, MPI message lengths, MPI reductions.
 */
#undef __FUNCT__
#define __FUNCT__ "PetscPrintXMLGlobalPerformanceElement"
static PetscErrorCode  PetscPrintXMLGlobalPerformanceElement(MPI_Comm comm,  FILE *fd, char *name, char *desc, PetscLogDouble max, PetscLogDouble ratio, PetscLogDouble avg, PetscLogDouble tot)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = PetscXMLStartSection(comm, fd, name, desc);CHKERRQ(ierr);
  ierr = PetscXMLPutDouble(comm, fd, "max", "", max, "%e");CHKERRQ(ierr);
  ierr = PetscXMLPutDouble(comm, fd, "ratio", "", ratio, "%f");CHKERRQ(ierr);
  if (avg>-1.0) {
    ierr = PetscXMLPutDouble(comm, fd, "average", "", avg, "%e");CHKERRQ(ierr);
  }
  if (tot>-1.0) {
    ierr = PetscXMLPutDouble(comm, fd, "total", "", tot, "%e");CHKERRQ(ierr);
  }
  ierr = PetscXMLEndSection(comm, fd, name);CHKERRQ(ierr);

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* Print the global performance: max, max/min, average and total of 
 *      time, objects, flops, flops/sec, memory, MPI messages, MPI message lengths, MPI reductions.
 */
#undef __FUNCT__
#define __FUNCT__ "PetscPrintGlobalPerformance"
static PetscErrorCode  PetscPrintGlobalPerformance(MPI_Comm comm,  FILE *fd, PetscLogDouble locTotalTime)
{
  PetscErrorCode     ierr;
  PetscLogDouble     min, max, tot, ratio, avg;
  PetscLogDouble     flops, mem, red, mess;
  PetscMPIInt        size;

  PetscFunctionBegin;
  ierr = MPI_Comm_size(comm, &size);CHKERRQ(ierr);

  /* Must preserve reduction count before we go on */
  red = petsc_allreduce_ct + petsc_gather_ct + petsc_scatter_ct;

  /* Calculate summary information */
  if (!useXMLFormat) {
    ierr = PetscFPrintf(comm, fd, "\n                          Max           Max/Min   Average    Total \n");CHKERRQ(ierr);
  } else {
    ierr = PetscXMLStartSection(comm, fd, "globalperformance", "Global performance");CHKERRQ(ierr);
  }
  /*   Time */
  ierr = MPI_Allreduce(&locTotalTime, &min, 1, MPIU_PETSCLOGDOUBLE, MPI_MIN, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&locTotalTime, &max, 1, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&locTotalTime, &tot, 1, MPIU_PETSCLOGDOUBLE, MPI_SUM, comm);CHKERRQ(ierr);
  avg  = (tot)/((PetscLogDouble) size);
  if (min != 0.0) ratio = max/min;
  else ratio = 0.0;
  if (!useXMLFormat) {
    ierr = PetscFPrintf(comm, fd, "Time (sec):               %5.3e  %10.5f   %5.3e\n", max, ratio, avg);CHKERRQ(ierr);
  } else {
    ierr = PetscPrintXMLGlobalPerformanceElement(comm, fd, "time", "Time (sec)", max, ratio, avg, -1.0);CHKERRQ(ierr);
  }
  /*   Objects */
  avg  = (PetscLogDouble) petsc_numObjects;
  ierr = MPI_Allreduce(&avg,          &min, 1, MPIU_PETSCLOGDOUBLE, MPI_MIN, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&avg,          &max, 1, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&avg,          &tot, 1, MPIU_PETSCLOGDOUBLE, MPI_SUM, comm);CHKERRQ(ierr);
  avg  = (tot)/((PetscLogDouble) size);
  if (min != 0.0) ratio = max/min;
  else ratio = 0.0;
  if (!useXMLFormat) {
    ierr = PetscFPrintf(comm, fd, "Objects:                  %5.3e  %10.5f   %5.3e\n", max, ratio, avg);CHKERRQ(ierr);
  } else {
    ierr = PetscPrintXMLGlobalPerformanceElement(comm, fd, "objects", "Objects", max, ratio, avg, -1.0);CHKERRQ(ierr);
  }
  /*   Flop */
  ierr = MPI_Allreduce(&petsc_TotalFlops,  &min, 1, MPIU_PETSCLOGDOUBLE, MPI_MIN, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&petsc_TotalFlops,  &max, 1, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&petsc_TotalFlops,  &tot, 1, MPIU_PETSCLOGDOUBLE, MPI_SUM, comm);CHKERRQ(ierr);
  avg  = (tot)/((PetscLogDouble) size);
  if (min != 0.0) ratio = max/min;
  else ratio = 0.0;
  if (!useXMLFormat) {
    ierr = PetscFPrintf(comm, fd, "MFlop:                    %5.3e  %10.5f   %5.3e  %5.3e\n", max/1.0E6, ratio, avg/1.0E6, tot/1.0E6);CHKERRQ(ierr);
  } else {
    ierr = PetscPrintXMLGlobalPerformanceElement(comm, fd, "mflop", "MFlop", max/1.0E6, ratio, avg/1.0E6, tot/1.0E6);CHKERRQ(ierr);
  }
  /*   Flop/sec -- Must talk to Barry here */
  if (locTotalTime != 0.0) flops = petsc_TotalFlops/locTotalTime;
  else flops = 0.0;
  ierr = MPI_Allreduce(&flops,        &min, 1, MPIU_PETSCLOGDOUBLE, MPI_MIN, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&flops,        &max, 1, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&flops,        &tot, 1, MPIU_PETSCLOGDOUBLE, MPI_SUM, comm);CHKERRQ(ierr);
  avg  = (tot)/((PetscLogDouble) size);
  if (min != 0.0) ratio = max/min;
  else ratio = 0.0;
  if (!useXMLFormat) {
    ierr = PetscFPrintf(comm, fd, "MFlop/sec:                %5.3e  %10.5f   %5.3e  %5.3e\n", max/1.0E6, ratio, avg/1.0E6, tot/1.0E6);CHKERRQ(ierr);
  } else {
    ierr = PetscPrintXMLGlobalPerformanceElement(comm, fd, "mflops", "MFlop/sec", max/1.0E6, ratio, avg/1.0E6, tot/1.0E6);CHKERRQ(ierr);
  }
  /*   Memory */
  ierr = PetscMallocGetMaximumUsage(&mem);CHKERRQ(ierr);
  if (mem > 0.0) {
    ierr = MPI_Allreduce(&mem,          &max, 1, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
    ierr = MPI_Allreduce(&mem,          &min, 1, MPIU_PETSCLOGDOUBLE, MPI_MIN, comm);CHKERRQ(ierr);
    ierr = MPI_Allreduce(&mem,          &tot, 1, MPIU_PETSCLOGDOUBLE, MPI_SUM, comm);CHKERRQ(ierr);
    avg  = (tot)/((PetscLogDouble) size);
    if (min != 0.0) ratio = max/min;
    else ratio = 0.0;
    if (!useXMLFormat) {
      ierr = PetscFPrintf(comm, fd, "Memory (MiB):             %5.3e  %10.5f   %5.3e  %5.3e\n", max/1024.0/1024.0, ratio, avg/1024.0/1024.0, tot/1024.0/1024.0);CHKERRQ(ierr);
    } else {
      ierr = PetscPrintXMLGlobalPerformanceElement(comm, fd, "memory", "Memory (MiB)", max/1024.0/1024.0, ratio, avg/1024.0/1024.0, tot/1024.0/1024.0);CHKERRQ(ierr);
    }
  }
  /*   Messages */
  mess = 0.5*(petsc_irecv_ct + petsc_isend_ct + petsc_recv_ct + petsc_send_ct);
  ierr = MPI_Allreduce(&mess,         &min, 1, MPIU_PETSCLOGDOUBLE, MPI_MIN, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&mess,         &max, 1, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&mess,         &tot, 1, MPIU_PETSCLOGDOUBLE, MPI_SUM, comm);CHKERRQ(ierr);
  avg  = (tot)/((PetscLogDouble) size);
  if (min != 0.0) ratio = max/min;
  else ratio = 0.0;
  if (!useXMLFormat) {
    ierr = PetscFPrintf(comm, fd, "MPI Message Transfers:    %5.3e  %10.5f   %5.3e  %5.3e\n", max, ratio, avg, tot);CHKERRQ(ierr);
  } else {
    ierr = PetscPrintXMLGlobalPerformanceElement(comm, fd, "messagetransfers", "MPI Message Transfers", max, ratio, avg, tot);CHKERRQ(ierr);
  }
  /*   Message Volume */
  mess = 0.5*(petsc_irecv_len + petsc_isend_len + petsc_recv_len + petsc_send_len);
  ierr = MPI_Allreduce(&mess,         &min, 1, MPIU_PETSCLOGDOUBLE, MPI_MIN, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&mess,         &max, 1, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&mess,         &tot, 1, MPIU_PETSCLOGDOUBLE, MPI_SUM, comm);CHKERRQ(ierr);
  avg = (tot)/((PetscLogDouble) size);
  if (min != 0.0) ratio = max/min;
  else ratio = 0.0;
  if (!useXMLFormat) {
    ierr = PetscFPrintf(comm, fd, "MPI Message Volume (MiB): %5.3e  %10.5f   %5.3e  %5.3e\n", max/1024.0/1024.0, ratio, avg/1024.0/1024.0, tot/1024.0/1024.0);CHKERRQ(ierr);
  } else {
    ierr = PetscPrintXMLGlobalPerformanceElement(comm, fd, "messagevolume", "MPI Message Volume (MiB)", max/1024.0/1024.0, ratio, avg/1024.0/1024.0, tot/1024.0/1024.0);CHKERRQ(ierr);
  }
  /*   Reductions */
  ierr = MPI_Allreduce(&red,          &min, 1, MPIU_PETSCLOGDOUBLE, MPI_MIN, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&red,          &max, 1, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
  ierr = MPI_Allreduce(&red,          &tot, 1, MPIU_PETSCLOGDOUBLE, MPI_SUM, comm);CHKERRQ(ierr);
  if (min != 0.0) ratio = max/min;
  else ratio = 0.0;
  if (!useXMLFormat) {
    ierr = PetscFPrintf(comm, fd, "MPI Reductions:           %5.3e  %10.5f\n", max, ratio);CHKERRQ(ierr);
  } else {
    ierr = PetscPrintXMLGlobalPerformanceElement(comm, fd, "reductions", "MPI Reductions", max, ratio, -1, -1);CHKERRQ(ierr);
  }
  if (!useXMLFormat) {
    ierr = PetscFPrintf(comm, fd, "\nFlop counting convention: 1 flop = 1 real number operation of type (multiply/divide/add/subtract)\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "                            e.g., VecAXPY() for real vectors of length N --> 2N flops\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "                            and VecAXPY() for complex vectors of length N --> 8N flops\n\n");CHKERRQ(ierr);
  } else {
    ierr = PetscXMLEndSection(comm, fd, "globalperformance");CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* If necessary, print a warning that this version of Petsc is likely not to be well optimized.
 */
#undef __FUNCT__
#define __FUNCT__ "PetscPrintPerformanceWarning"
static PetscErrorCode  PetscPrintPerformanceWarning(MPI_Comm comm,  FILE *fd)
{
  PetscErrorCode     ierr;
  PetscFunctionBegin;
  if(!useXMLFormat) {
#if defined(PETSC_USE_DEBUG)
    ierr = PetscFPrintf(comm, fd, "\n\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      ##########################################################\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #                                                        #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #                          WARNING!!!                    #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #                                                        #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #   This code was compiled with a debugging option,      #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #   To get timing results run ./configure                #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #   using --with-debugging=no, the performance will      #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #   be generally two or three times faster.              #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #                                                        #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      ##########################################################\n\n\n");CHKERRQ(ierr);
#endif
#if defined(PETSC_USE_COMPLEX) && !defined(PETSC_USE_FORTRAN_KERNELS)
    ierr = PetscFPrintf(comm, fd, "\n\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      ##########################################################\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #                                                        #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #                          WARNING!!!                    #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #                                                        #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #   The code for various complex numbers numerical       #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #   kernels uses C++, which generally is not well        #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #   optimized.  For performance that is about 4-5 times  #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #   faster, specify --with-fortran-kernels=1             #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #   when running ./configure.py.                         #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      #                                                        #\n");CHKERRQ(ierr);
    ierr = PetscFPrintf(comm, fd, "      ##########################################################\n\n\n");CHKERRQ(ierr);
#endif
  }
  PetscFunctionReturn(0);
}


/******************************************************************************************/
typedef struct {
  PetscLogEvent  dftEvent;
  NestedEventId  nstEvent;
  PetscLogEvent  dftParent;
  NestedEventId  nstParent;
  PetscBool      own;
  size_t         depth;
  NestedEventId* nstPath;
} PetscNestedEventTree;

/******************************************************************************************/
/* Compare timers to sort them in the tree */
static int compareTreeItems(const void *item1_, const void *item2_) {
  int i;
  PetscNestedEventTree *item1 = (PetscNestedEventTree *) item1_;
  PetscNestedEventTree *item2 = (PetscNestedEventTree *) item2_;
  for (i=0; i<MIN(item1->depth,item2->depth); i++) {
     if (item1->nstPath[i]<item2->nstPath[i]) return -1;
     if (item1->nstPath[i]>item2->nstPath[i]) return +1;
  }
  if (item1->depth < item2->depth) return -1;
  if (item1->depth > item2->depth) return 1;
  return 0;
}
/************************************************************************************************
 * Do MPI communication to get the complete, nested calling tree for all processes: there may be
 * calls that happen in some processes, but not in others.
 *
 * The output, tree[nTimers] is an array of PetscNestedEventTree-structs.
 * The tree is sorted so that the timers can be printed in the order of appearance.
 *
 * For tree-items which appear in the trees of multiple processes (which will be most items), the 
 * following rule is followed:
 * + if information from my own process is available, then that is the information stored in tree.
 *   otherwise it is some other process's information.
 */
#undef __FUNCT__
#define __FUNCT__ "PetscCreateLogTreeNested"
static PetscErrorCode  PetscCreateLogTreeNested(MPI_Comm comm, FILE *fd, PetscNestedEventTree **p_tree, unsigned long long *p_nTimers)
{
  PetscNestedEventTree *tree, *newTree;
  size_t *treeIndices;
  int                nTimers, totalNTimers, i, j, iTimer0, maxDefaultTimer;
  int                yesno;
  PetscBool          done;
  PetscMPIInt        size,rank;
  PetscErrorCode     ierr;
  int                maxdepth;
  int                depth;
  int                illegalEvent;
  size_t             iextra;
  PetscStageLog      stageLog;
  PetscEventRegInfo  *eventInfo; 
  NestedEventId      *nstPath, *nstMyPath;

  PetscFunctionBegin;
  ierr = MPI_Comm_rank(comm, &rank);CHKERRQ(ierr);
  ierr = MPI_Comm_size(comm, &size);CHKERRQ(ierr);
  ierr = PetscLogGetStageLog(&stageLog);CHKERRQ(ierr);
  eventInfo = stageLog->eventLog->eventInfo;

  /* Calculate memory needed to store everybody's information and allocate tree */
  nTimers = 0;
  for (i=0; i<nNestedEvents; i++) nTimers+=nestedEvents[i].nParents;

  ierr = PetscMalloc(nTimers*sizeof(PetscNestedEventTree),&tree); CHKERRQ(ierr);

  /* Fill tree with readily available information */ 
  iTimer0 = 0;
  maxDefaultTimer =0;
  for (i=0; i<nNestedEvents; i++) {
     size_t nParents                 = nestedEvents[i].nParents;
     NestedEventId nstEvent          = nestedEvents[i].nstEvent;
     PetscLogEvent *dftParentsSorted = nestedEvents[i].dftParentsSorted;
     PetscLogEvent *dftEvents        = nestedEvents[i].dftEvents;
     for (j=0; j<nParents; j++) {
         maxDefaultTimer = MAX(dftEvents[j],maxDefaultTimer);

         tree[iTimer0+j].dftEvent   = dftEvents[j];
         tree[iTimer0+j].nstEvent   = nstEvent;
         tree[iTimer0+j].dftParent  = dftParentsSorted[j];
         tree[iTimer0+j].own        = PETSC_TRUE;

         tree[iTimer0+j].nstParent  = 0;
         tree[iTimer0+j].depth      = 0;
         tree[iTimer0+j].nstPath    = NULL;
     }
     iTimer0 += nParents;
  }

# if VERBOSE 
     if (rank<=1) {
        for (i = 0; i<=maxDefaultTimer; i++) {
           printf("[%d] default timer %d : '%s'\n", rank, (int) i, eventInfo[i].name);
        }
     }
# endif

  /* Calculate the global maximum for the default timer index, so array treeIndices can 
   * be allocated only once */
  ierr = MPI_Allreduce(&maxDefaultTimer, &j, 1, MPI_INT, MPI_MAX, comm);CHKERRQ(ierr);
  maxDefaultTimer = j;

  /* Find default timer's place in the tree */
  ierr = PetscMalloc((maxDefaultTimer+1)*sizeof(size_t),&treeIndices); CHKERRQ(ierr);
  treeIndices[0] = 0;
  for (i=1; i<nTimers; i++) {
      PetscLogEvent dftEvent = tree[i].dftEvent;
      treeIndices[dftEvent] = i;
  }

  /* Find each dftParent's nested identification */
  for (i=0; i<nTimers; i++) {
      PetscLogEvent dftParent = tree[i].dftParent;
      if (dftParent!=0) {
         size_t j = treeIndices[dftParent];
         tree[i].nstParent  = tree[j].nstEvent;
      }
  }

  /* Find depths for each timer path */
  done = PETSC_FALSE;
  maxdepth = 0;
  tree[0].depth=1;
  while (!done) {
     done = PETSC_TRUE;
     for (i=1; i<nTimers; i++) {
         int j = treeIndices[tree[i].dftParent];
         if (j==0) {
            tree[i].depth=1;
         } else if (tree[i].dftEvent!=0) {
            depth = 1+tree[j].depth;
            if (depth>tree[i].depth) {
               done          = PETSC_FALSE;
               tree[i].depth = depth;
               maxdepth      = MAX(depth,maxdepth);
            }
         }
     }
  }

  /* Allocate the paths in the entire tree */
  for (i=0; i<nTimers; i++) {
      depth = tree[i].depth;
      ierr = PetscMalloc(depth*sizeof(NestedEventId),&tree[i].nstPath); CHKERRQ(ierr);
  }

  /* Calculate the paths for all timers */
  for (depth=1; depth<=maxdepth; depth++) {
      for (i=0; i<nTimers; i++) {
          if (tree[i].depth==depth) {
             if (depth>1) {
                size_t j = treeIndices[tree[i].dftParent];
                memcpy(tree[i].nstPath,tree[j].nstPath,(depth-1)*sizeof(NestedEventId));
             }
             tree[i].nstPath[depth-1] = tree[i].nstEvent;
          }
      }
  }
  ierr = PetscFree(treeIndices);CHKERRQ(ierr);

  /* Sort the tree on basis of the paths */
  qsort(tree, nTimers, sizeof(PetscNestedEventTree), compareTreeItems);

  /* Allocate an array to store paths */
  depth = maxdepth;
  ierr = MPI_Allreduce(&depth, &maxdepth, 1, MPI_INT, MPI_MAX, comm);CHKERRQ(ierr);
  ierr = PetscMalloc((maxdepth+1)*sizeof(NestedEventId), &nstPath); CHKERRQ(ierr);
  ierr = PetscMalloc((maxdepth+1)*sizeof(NestedEventId), &nstMyPath); CHKERRQ(ierr);

  /* Find an illegal nested event index (1+largest nested event index) */
  illegalEvent = 1+nestedEvents[nNestedEvents-1].nstEvent;
  i = illegalEvent;
  ierr = MPI_Allreduce(&i, &illegalEvent, 1, MPI_INT, MPI_MAX, comm);CHKERRQ(ierr);

  /* First, detect timers which are not available in this process, but are available in others
   *        Allocate a new tree, that can contain all timers
   * Then,  fill the new tree with all (own and not-own) timers */
  newTree= NULL;
  for (yesno=0; yesno<=1; yesno++) {
     depth  = 1;
     i      = 0;
     iextra = 0;
     while (depth>0) {
        size_t    j;
        PetscBool same;

        /* Construct the next path in this process's tree:
         * if necessary, supplement with invalid path entries */
        depth++;
        if (i<nTimers) {
           for (j=0;             j<tree[i].depth; j++) nstMyPath[j] = tree[i].nstPath[j];
           for (j=tree[i].depth; j<depth;         j++) nstMyPath[j] = illegalEvent;
        } else {
           for (j=0;             j<depth;         j++) nstMyPath[j] = illegalEvent;
        }
    
        /* Communicate with other processes to obtain the next path and its depth */
        ierr = MPI_Allreduce(nstMyPath, nstPath, depth, MPI_INT, MPI_MIN, comm);CHKERRQ(ierr);
        for (j=depth-1; (int) j>=0; j--) { 
            if (nstPath[j]==illegalEvent) depth=j;
        }
    
        if (depth>0) {
        /* If the path exists */

           /* check whether the next path is the same as this process's next path */
           same = PETSC_TRUE;     
           for (j=0; same && j<=depth; j++) { same = (PetscBool) (same &&  nstMyPath[j] == nstPath[j]);}
          
           if (same) { 
           /* Register 'own path' */
               if (newTree) newTree[i+iextra] = tree[i];
               i++;  
           } else { 
           /* Register 'not an own path' */
               if (newTree) { 
                   newTree[i+iextra].nstEvent   = nstPath[depth-1];
                   newTree[i+iextra].own        = PETSC_FALSE;
                   newTree[i+iextra].depth      = depth;
                   ierr = PetscMalloc(depth*sizeof(NestedEventId), &newTree[i+iextra].nstPath); CHKERRQ(ierr);
                   for (j=0; j<depth; j++) {newTree[i+iextra].nstPath[j] = nstPath[j];}
          
                   newTree[i+iextra].dftEvent  = 0;
                   newTree[i+iextra].dftParent = 0;
                   newTree[i+iextra].nstParent = 0;
               }
               iextra++; 
           }

           /* Print the current path */
#          if VERBOSE
              ierr = PetscFPrintf(comm, fd, "%s", newTree ? " >" : "  ");CHKERRQ(ierr);
              for (j=0; j<depth; j++) {
                  ierr = PetscFPrintf(comm, fd, "> %s ",eventInfo[(PetscLogEvent) nstPath[j]].name);CHKERRQ(ierr);
              }
              ierr = PetscFPrintf(comm, fd, "\n");CHKERRQ(ierr);
#          endif
        }
     }

     /* Determine the size of the complete tree (with own and not-own timers) and allocate the new
      * tree */
     totalNTimers = nTimers + iextra;
     if (!newTree) {
         ierr = PetscMalloc(totalNTimers*sizeof(PetscNestedEventTree), &newTree); CHKERRQ(ierr);
     }

  }

  ierr = PetscFree(nstPath);CHKERRQ(ierr);
  ierr = PetscFree(nstMyPath);CHKERRQ(ierr);
  ierr = PetscFree(tree);CHKERRQ(ierr);
  tree = newTree;
  newTree = NULL;

# if VERBOSE  
     for (i=0; i<totalNTimers; i++) {
         for (j=0; j<tree[i].depth; j++) {
             NestedEventId nstEvent = tree[i].nstPath[j];
             const char *name = eventInfo[(PetscLogEvent) nstEvent].name;
             ierr = PetscFPrintf(comm, fd, " > %s (%d) ",name, (int) nstEvent);CHKERRQ(ierr);
         }
         ierr = PetscFPrintf(comm, fd, "\n");CHKERRQ(ierr);
     }
# endif

  /* Set return value and return */
  *p_tree    = tree;
  *p_nTimers = totalNTimers;

  PetscFunctionReturn(0);
}

/************************************************************************************************
 * Delete the nested timer tree 
 */
#undef __FUNCT__
#define __FUNCT__ "PetscLogFreeNestedTree"
static PetscErrorCode  PetscLogFreeNestedTree(PetscNestedEventTree *tree, const unsigned long long nTimers)
{
  unsigned long long i;
  PetscErrorCode     ierr;
  PetscFunctionBegin;
  for (i=0; i<nTimers; i++) {
      ierr = PetscFree(tree[i].nstPath);CHKERRQ(ierr);
  }
  ierr = PetscFree(tree);CHKERRQ(ierr);

  PetscFunctionReturn(0);
}


/******************************************************************************************/
/* Print the global performance: max, max/min, average and total of 
 *      time, objects, flops, flops/sec, memory, MPI messages, MPI message lengths, MPI reductions.
 */
#undef __FUNCT__
#define __FUNCT__ "PetscPrintXMLNestedLinePerfResults"
static PetscErrorCode  PetscPrintXMLNestedLinePerfResults(MPI_Comm comm,  FILE *fd, char *name, PetscLogDouble minvalue, PetscLogDouble maxvalue, PetscLogDouble minmaxtreshold)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = PetscXMLStartSection(comm, fd, name, "");CHKERRQ(ierr);
  if (maxvalue>minvalue*minmaxtreshold) {
    ierr = PetscXMLPutDouble(comm, fd, "minvalue", "", minvalue, "%f"); CHKERRQ(ierr);
    ierr = PetscXMLPutDouble(comm, fd, "maxvalue", "", maxvalue, "%f"); CHKERRQ(ierr);
  } else {
    ierr = PetscXMLPutDouble(comm, fd, "value", "", (minvalue+maxvalue)/2.0, "%f"); CHKERRQ(ierr);
  };
  ierr = PetscXMLEndSection(comm, fd, name);CHKERRQ(ierr);

  PetscFunctionReturn(0);
}


/************************************************************************************************
 * Print one line in performance tree
 */
#undef __FUNCT__
#define __FUNCT__ "PetscLogPrintNestedLine"
static PetscErrorCode  PetscLogPrintNestedLine(
      MPI_Comm comm, FILE *fd, 
      PetscEventPerfInfo perfInfo, 
      PetscLogDouble countsPerCall, 
      int parentCount, 
      size_t depth, 
      const char *name,
      PetscLogDouble totalTime, 
      bool *isPrinted
) {
  const int debug=0;
  PetscLogDouble time = perfInfo.time;

  PetscLogDouble timeMx,          timeMn;
  PetscLogDouble countsPerCallMx, countsPerCallMn;
  PetscLogDouble reductSpeedMx,   reductSpeedMn;
  PetscLogDouble flopSpeedMx,     flopSpeedMn;
  PetscLogDouble msgSpeedMx,      msgSpeedMn;

  const int n_comm = 8;
  PetscLogDouble commarr_in[n_comm], commarr_out[n_comm];

  PetscErrorCode ierr;
  size_t         j;
  PetscMPIInt        rank;

  PetscFunctionBegin;
  ierr = MPI_Comm_rank(comm, &rank);CHKERRQ(ierr);

  if (debug>=1) {
    printf("[%d]   collecting info for %s (%d)\n", rank, name, (int) depth);
    printf("[%d]   my %s: time = %e\n", rank, name, time);
  }

  commarr_in[0] =  time; 
  commarr_in[1] = -time; 
  ierr = MPI_Allreduce(commarr_in, commarr_out,    2, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
  timeMx =  commarr_out[0];
  timeMn = -commarr_out[1];

  if (debug>=1) {
    printf("[%d]    time for %s (%d): %e-%e\n", rank, name, (int) depth, timeMx, timeMn);
  }

  commarr_in[0] = time>=timeMx*0.9 ?  perfInfo.flops/time         : 0;
  commarr_in[1] = time>=timeMx*0.9 ?  perfInfo.numReductions/time : 0;
  commarr_in[2] = time>=timeMx*0.9 ?  perfInfo.messageLength/time : 0;
  commarr_in[3] = parentCount>0    ?  countsPerCall      : 0;

  commarr_in[4] = time>=timeMx*0.9 ? -perfInfo.flops/time         : -1e30;
  commarr_in[5] = time>=timeMx*0.9 ? -perfInfo.numReductions/time : -1e30;
  commarr_in[6] = time>=timeMx*0.9 ? -perfInfo.messageLength/time : -1e30;
  commarr_in[7] = parentCount>0    ? -countsPerCall      : -1e30;

  ierr = MPI_Allreduce(commarr_in, commarr_out,  n_comm, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);

  flopSpeedMx     =  commarr_out[0];
  reductSpeedMx   =  commarr_out[1];
  msgSpeedMx      =  commarr_out[2];
  countsPerCallMx =  commarr_out[3];

  flopSpeedMn     = -commarr_out[4];
  reductSpeedMn   = -commarr_out[5];
  msgSpeedMn      = -commarr_out[6];
  countsPerCallMn = -commarr_out[7];

  *isPrinted = ((timeMx/totalTime) > (threshTime/100.0));
  if (isPrinted) {

    if (!useXMLFormat) {
      for (j=0; j<depth; j++) { 
        ierr = PetscFPrintf(comm, fd, "  ");CHKERRQ(ierr);
      }
      ierr = PetscFPrintf(comm, fd, "%-25s ",name);CHKERRQ(ierr);
      if (timeMx>timeMn*1.02) {
        ierr = PetscFPrintf(comm, fd, "%0.0f - %0.0f %%",timeMn/totalTime*100.0,timeMx/totalTime*100.0);CHKERRQ(ierr);
      } else {
          ierr = PetscFPrintf(comm, fd, "%0.0f %%",(timeMn+timeMx)/2.0/totalTime*100.0);CHKERRQ(ierr);
      };
 
      if (countsPerCallMx<1.01 && countsPerCallMn>0.99) {
          /* One call per parent */
      } else if (countsPerCallMx > 1.02 * countsPerCallMn) {
          ierr = PetscFPrintf(comm, fd, "   (%0.1f - %-0.1f calls/parent call)",countsPerCallMn,countsPerCallMx);CHKERRQ(ierr);
      } else {
          ierr = PetscFPrintf(comm, fd, "   (%0.1f calls/parent call)",(countsPerCallMx+countsPerCallMn)/2);CHKERRQ(ierr);
      }
 
      if (flopSpeedMx<0.01) {
          /* NO flops: don't print */
      } else if (flopSpeedMx>flopSpeedMn*1.05) {
          ierr = PetscFPrintf(comm, fd, "   {%0.2f - %-0.2f Mflop/sec}",flopSpeedMn/1e6,flopSpeedMx/1e6);CHKERRQ(ierr);
      } else {
          ierr = PetscFPrintf(comm, fd, "   {%0.2f Mflop/sec}",(flopSpeedMn+flopSpeedMx)/2e6);CHKERRQ(ierr);
      };
 
      if (msgSpeedMx<0.01) {
          /* NO msgs: don't print */
      } else if (msgSpeedMx>msgSpeedMn*1.05) {
          ierr = PetscFPrintf(comm, fd, "   [%0.2f - %-0.2f MBytes/sec]",msgSpeedMn/1024/1024,msgSpeedMx/1024/1024);CHKERRQ(ierr);
      } else {
          ierr = PetscFPrintf(comm, fd, "   [%0.2f MBytes/sec]",(msgSpeedMn+msgSpeedMx)/2/1024/1024);CHKERRQ(ierr);
      };
 
      if (reductSpeedMx<0.01) {
          /* NO reductions: don't print */
      } else if (reductSpeedMx>reductSpeedMn*1.05) {
          ierr = PetscFPrintf(comm, fd, "   <%0.2f - %-0.2f Reductions/sec>",reductSpeedMn,reductSpeedMx);CHKERRQ(ierr);
      } else {
          ierr = PetscFPrintf(comm, fd, "   <%0.2f Reductions/sec>",(reductSpeedMn+reductSpeedMx)/2);CHKERRQ(ierr);
      };
 
      ierr = PetscFPrintf(comm, fd, "\n");CHKERRQ(ierr);
    } else {
      
      ierr = PetscXMLStartSection(comm, fd, "event", ""); CHKERRQ(ierr);
      ierr = PetscXMLPutString(comm, fd, "name", "", name); CHKERRQ(ierr);

      ierr = PetscPrintXMLNestedLinePerfResults(comm, fd, "time", timeMn/totalTime*100.0, timeMx/totalTime*100.0, 1.02); CHKERRQ(ierr);
 
      if (countsPerCallMx<1.01 && countsPerCallMn>0.99) {
          /* One call per parent */
      } else {
        ierr = PetscPrintXMLNestedLinePerfResults(comm, fd, "ncalls", countsPerCallMn, countsPerCallMx, 1.02); CHKERRQ(ierr);
      }
 
      if (flopSpeedMx<0.01) {
          /* NO flops: don't print */
      } else {
        ierr = PetscPrintXMLNestedLinePerfResults(comm, fd, "mflops", flopSpeedMn/1e6, flopSpeedMx/1e6, 1.05); CHKERRQ(ierr);
      }
 
      if (msgSpeedMx<0.01) {
          /* NO msgs: don't print */
      } else {
        ierr = PetscPrintXMLNestedLinePerfResults(comm, fd, "mbps", msgSpeedMn/1024.0/1024.0, msgSpeedMx/1024.0/1024.0, 1.05); CHKERRQ(ierr);
      }
 
      if (reductSpeedMx<0.01) {
          /* NO reductions: don't print */
      } else {
        ierr = PetscPrintXMLNestedLinePerfResults(comm, fd, "nreductsps", reductSpeedMn, reductSpeedMx, 1.05); CHKERRQ(ierr);
      }

      //PetscXMLEndSection(comm,fd, "event") has to be done in the calling routine; in case of
      //children, an "events" field has to be added to this "event" section
 
    }
  }
  if (debug>=1) {
     printf("[%d]   done collecting info for %s (%d)\n", rank, name, (int) depth);
  }
  PetscFunctionReturn(0);
}

/************************************************************************************************
 * Count the number of times the parent event was called
 */
  
static int countParents( const PetscNestedEventTree *tree, PetscEventPerfInfo *eventPerfInfo, const int i) {  
   if (tree[i].depth<=1) {
      return 1;  /* Main event: only once */
   } else if (!tree[i].own) {
      return 1;  /* This event didn't happen in this process, but did in another */
   } else {
      int iParent;
      for (iParent=i-1; iParent>=0; iParent--) {
         if (tree[iParent].depth == tree[i].depth-1) break;
      }
      if (tree[iParent].depth != tree[i].depth-1) {
         printf("\n\n   *****  Internal error: cannot find parent ****\n\n");
         return -2;
      } else {
         PetscLogEvent dftEvent  = tree[iParent].dftEvent;
         return eventPerfInfo[dftEvent].count;
      }
   }
}

/************************************************************************************************
 */
typedef struct {
  int             id; 
  PetscLogDouble  val;                                  
} PetscSortItem;

static int compareSortItems(const void *item1_, const void *item2_) {
  int i;
  PetscSortItem *item1 = (PetscSortItem *) item1_;
  PetscSortItem *item2 = (PetscSortItem *) item2_;
  if (item1->val > item2->val) return -1;
  if (item1->val < item2->val) return +1;
  return 0;
}

static PetscErrorCode  PetscLogNestedPrint(MPI_Comm comm, FILE *fd, PetscNestedEventTree *tree, 
                            const unsigned long long nTimers, const unsigned long long iStart, PetscLogDouble totalTime)
{
  size_t             depth   = tree[iStart].depth; 
  NestedEventId*     nstPath = tree[iStart].nstPath;
  const char         *name;
  unsigned long long nChildren;
  PetscSortItem      *children;
  PetscErrorCode     ierr;
  PetscEventPerfInfo *eventPerfInfo;
  PetscEventPerfInfo myPerfInfo,  otherPerfInfo, selfPerfInfo;
  PetscLogDouble     parentCount, countsPerCall;
  const int          debug=0;
  bool               wasPrinted;
  bool               childWasPrinted;

  {
  /* Look up the name of the event and its PerfInfo */
     const int          stage=0;
     PetscStageLog      stageLog;
     PetscEventRegInfo  *eventRegInfo; 
     ierr          = PetscLogGetStageLog(&stageLog);CHKERRQ(ierr);
     eventRegInfo  = stageLog->eventLog->eventInfo;
     eventPerfInfo = stageLog->stageInfo[stage].eventLog->eventInfo;
     name          = eventRegInfo[(PetscLogEvent) tree[iStart].nstEvent].name;
  }

  if ((debug>=1)&&(!useXMLFormat)) {
     ierr = PetscFPrintf(comm, fd, " printing tree for %s\n",name); CHKERRQ(ierr);
  }
  
  /* Count the number of child processes */
  nChildren = 0;
  { 
     int i;
     for (i=iStart+1; i<nTimers; i++) {
         if (tree[i].depth<=depth) break;
         if (tree[i].depth == depth + 1) nChildren++;
     }
  }

  if (nChildren>0) {
     /* Create an array for the id-s and maxTimes of the children,
      *  leaving 2 spaces for self-time and other-time */
     int i;
     PetscLogDouble *times, *maxTimes;

     ierr = PetscMalloc( (nChildren+2)*sizeof(PetscSortItem),&children);CHKERRQ(ierr);
     nChildren = 0;
     for (i=iStart+1; i<nTimers; i++) {
         if (tree[i].depth<=depth) break;
         if (tree[i].depth == depth + 1) {
             children[nChildren].id  = i;
             children[nChildren].val = eventPerfInfo[tree[i].dftEvent].time ;
             nChildren++;
         }
     }

     /* Calculate the children's maximum times, to see whether children will be ignored or printed */
     ierr = PetscMalloc(nChildren*sizeof(PetscLogDouble),&times); CHKERRQ(ierr);
     for (i=0; i<nChildren; i++) { times[i] = children[i].val; }

     ierr = PetscMalloc(nChildren*sizeof(PetscLogDouble),&maxTimes); CHKERRQ(ierr);
     ierr = MPI_Allreduce(times, maxTimes, nChildren, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
     ierr = PetscFree(times); CHKERRQ(ierr);

     for (i=0; i<nChildren; i++) { children[i].val = maxTimes[i]; }
     ierr = PetscFree(maxTimes); CHKERRQ(ierr);
  }

  if (!tree[iStart].own) {
  /* Set values for a timer that was not activated in this process 
   * (but was, in other processes of this run) */
      myPerfInfo.time          = 0; 
      myPerfInfo.flops         = 0;
      myPerfInfo.numMessages   = 0;
      myPerfInfo.messageLength = 0;
      myPerfInfo.numReductions = 0;

      selfPerfInfo  = myPerfInfo;
      otherPerfInfo = myPerfInfo;

      parentCount   = 1;
      countsPerCall = 0;

  } else {
  /* Set the values for a timer that was activated in this process */
      int i;
      PetscLogEvent dftEvent   = tree[iStart].dftEvent;

      parentCount    = countParents( tree, eventPerfInfo, iStart);
      myPerfInfo     = eventPerfInfo[dftEvent];
      countsPerCall  = (PetscLogDouble) myPerfInfo.count / (PetscLogDouble) parentCount;

      selfPerfInfo                = myPerfInfo;
      otherPerfInfo.time          = 0; 
      otherPerfInfo.flops         = 0;
      otherPerfInfo.numMessages   = 0;
      otherPerfInfo.messageLength = 0;
      otherPerfInfo.numReductions = 0;

      for (i=0; i<nChildren; i++) {
      /* For all child counters: subtract the child values from self-timers */
       
          PetscLogEvent dftChild  = tree[children[i].id].dftEvent;
          PetscEventPerfInfo childPerfInfo = eventPerfInfo[dftChild];
          PetscLogDouble childTime;


          selfPerfInfo.time          -= childPerfInfo.time;
          selfPerfInfo.flops         -= childPerfInfo.flops;
          selfPerfInfo.numMessages   -= childPerfInfo.numMessages;
          selfPerfInfo.messageLength -= childPerfInfo.messageLength;
          selfPerfInfo.numReductions -= childPerfInfo.numReductions;

          if ((children[i].val/totalTime) < (threshTime/100.0)) {
          /* Add them to 'other' if the time is ignored in the output */
              if ((debug>=1)&&(!useXMLFormat)) {
                 ierr = PetscFPrintf(comm, fd, "    ignoring child %d (%e sec)\n",children[i].id, children[i].val); CHKERRQ(ierr);
              }
              otherPerfInfo.time          += childPerfInfo.time;
              otherPerfInfo.flops         += childPerfInfo.flops;
              otherPerfInfo.numMessages   += childPerfInfo.numMessages;
              otherPerfInfo.messageLength += childPerfInfo.messageLength;
              otherPerfInfo.numReductions += childPerfInfo.numReductions;
          }
      }
  }

  /* Main output for this timer */
  ierr = PetscLogPrintNestedLine(comm, fd, myPerfInfo, countsPerCall, parentCount, depth, name, totalTime, &wasPrinted); CHKERRQ(ierr); 

  /* Now print the lines for the children */
  if (nChildren>0) {
     /* Calculate max-times for 'self' and 'other' */
     int i;
     PetscLogDouble times[2], maxTimes[2];
     times[0] = selfPerfInfo.time;   times[1] = otherPerfInfo.time;
     ierr = MPI_Allreduce(times, maxTimes, 2, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
     children[nChildren+0].id = -1;
     children[nChildren+0].val = maxTimes[0];
     children[nChildren+1].id = -2;
     children[nChildren+1].val = maxTimes[1];

     /* Now sort the children (including 'self' and 'other') on total time */
     qsort(children, nChildren+2, sizeof(PetscSortItem), compareSortItems);

     /* Print (or ignore) the children in ascending order of total time */
     if (useXMLFormat) {
       ierr = PetscXMLStartSection(comm, fd,"events", ""); CHKERRQ(ierr);
     }
     for (i=0; i<nChildren+2; i++) {
         if ((children[i].val/totalTime) < (threshTime/100.0)) {
         /* ignored: no output */
         } else if (children[i].id==-1) {
         /* output for 'self' */
            char selfname[6+strlen(name)];
            sprintf(selfname,"self-%s",name);
            ierr = PetscLogPrintNestedLine(comm, fd, selfPerfInfo, countsPerCall, parentCount, depth+1, selfname, totalTime, &childWasPrinted); CHKERRQ(ierr); 
            if ((useXMLFormat)&&(childWasPrinted)) {
              ierr = PetscXMLEndSection(comm, fd,"event"); CHKERRQ(ierr);
            }
         } else if (children[i].id==-2) {
         /* output for 'other' */
            char otherName[15+strlen(name)];
            sprintf(otherName,"%s: other-timed",name);
            ierr = PetscLogPrintNestedLine(comm, fd, otherPerfInfo, 1, 1, depth+1, otherName, totalTime, &childWasPrinted); CHKERRQ(ierr); 
            if ((useXMLFormat)&&(childWasPrinted)) {
              ierr = PetscXMLEndSection(comm, fd,"event"); CHKERRQ(ierr);
            }
         } else {
         /* Print the child with a recursive call to this function */
            ierr = PetscLogNestedPrint(comm, fd, tree, nTimers, children[i].id, totalTime);
         }
     }
     if (useXMLFormat) {
       ierr = PetscXMLEndSection(comm, fd,"events"); CHKERRQ(ierr);
     }
     ierr = PetscFree(children); CHKERRQ(ierr);
  }

  if ((useXMLFormat)&&(wasPrinted)) {
    ierr = PetscXMLEndSection(comm, fd, "event"); CHKERRQ(ierr);
  }
  return 0;
}


/****************************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscLogNestedPrintTop"
static PetscErrorCode  PetscLogNestedPrintTop(MPI_Comm comm, FILE *fd, PetscNestedEventTree *tree, 
                            const unsigned long long nTimers, PetscLogDouble totalTime)
{
  unsigned long long nChildren;
  PetscSortItem      *children;
  PetscErrorCode     ierr;
  PetscEventPerfInfo *eventPerfInfo;

  PetscFunctionBegin;
  {
  /* Look up the PerfInfo */
     const int          stage=0;
     PetscStageLog      stageLog;
     PetscEventRegInfo  *eventRegInfo; 
     ierr          = PetscLogGetStageLog(&stageLog);CHKERRQ(ierr);
     eventRegInfo  = stageLog->eventLog->eventInfo;
     eventPerfInfo = stageLog->stageInfo[stage].eventLog->eventInfo;
  }

  /* Count the number of child processes, and count total time */
  nChildren = 0;
  { 
     int i;
     for (i=0; i<nTimers; i++) {
         if (tree[i].depth==1) nChildren++;
     }
  }

  if (nChildren>0) {
     /* Create an array for the id-s and maxTimes of the children,
      *  leaving 2 spaces for self-time and other-time */
     int i;
     PetscLogDouble *times, *maxTimes;

     ierr = PetscMalloc( nChildren*sizeof(PetscSortItem),&children);CHKERRQ(ierr);
     nChildren = 0;
     for (i=0; i<nTimers; i++) {
         if (tree[i].depth == 1) {
             children[nChildren].id  = i;
             children[nChildren].val = eventPerfInfo[tree[i].dftEvent].time ;
             nChildren++;
         }
     }

     /* Calculate the children's maximum times, to sort them */
     ierr = PetscMalloc(nChildren*sizeof(PetscLogDouble),&times); CHKERRQ(ierr);
     for (i=0; i<nChildren; i++) { times[i] = children[i].val; }

     ierr = PetscMalloc(nChildren*sizeof(PetscLogDouble),&maxTimes); CHKERRQ(ierr);
     ierr = MPI_Allreduce(times, maxTimes, nChildren, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
     ierr = PetscFree(times); CHKERRQ(ierr);

     for (i=0; i<nChildren; i++) { children[i].val = maxTimes[i]; }
     ierr = PetscFree(maxTimes); CHKERRQ(ierr);

     /* Now sort the children on total time */
     qsort(children, nChildren, sizeof(PetscSortItem), compareSortItems);
     /* Print (or ignore) the children in ascending order of total time */
     if (!useXMLFormat) {
        ierr = PetscFPrintf(comm, fd, "\n\nTimer tree (time is in %% of %0.2f s, threshold is %0.2f %%):\n\n", totalTime, threshTime); CHKERRQ(ierr);
     } else {
        ierr = PetscXMLStartSection(comm, fd, "timertree", "Timings tree"); CHKERRQ(ierr);
        ierr = PetscXMLPutDouble(comm, fd, "totaltime", "", totalTime, "%f");CHKERRQ(ierr);
        ierr = PetscXMLPutDouble(comm, fd, "timethreshold", "", threshTime, "%f");CHKERRQ(ierr);
     }
     for (i=0; i<nChildren; i++) {
         if ((children[i].val/totalTime) < (threshTime/100.0)) {
         /* ignored: no output */
         } else {
         /* Print the child with a recursive call to this function */
            ierr = PetscLogNestedPrint(comm, fd, tree, nTimers, children[i].id, totalTime);
         }
     }
     if (useXMLFormat) {
       ierr = PetscXMLEndSection(comm, fd, "timertree"); CHKERRQ(ierr);
     }
     ierr = PetscFree(children); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}
/*****************************************************************/
typedef struct {
  char           *name;
  PetscLogDouble time;
  PetscLogDouble flops;
  PetscLogDouble numMessages;
  PetscLogDouble messageLength;
  PetscLogDouble numReductions;
} PetscSelfTimer;
/*****************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscCalcSelfTime"
static PetscErrorCode  PetscCalcSelfTime(MPI_Comm comm, FILE *fd, PetscSelfTimer **p_self, unsigned long long *p_nstMax)
{
  PetscMPIInt        size,rank;
  PetscErrorCode     ierr;
  PetscEventPerfInfo *eventPerfInfo;
  PetscEventRegInfo  *eventRegInfo; 
  PetscSelfTimer     *selftimes;
  PetscSelfTimer     *totaltimes;
  NestedEventId      *nstEvents;
  unsigned long long i, maxDefaultTimer;
  NestedEventId      nst;
  PetscLogEvent      dft;
  int nstMax, nstMax_local;
  
  PetscFunctionBegin;

  ierr = MPI_Comm_rank(comm, &rank);CHKERRQ(ierr);
  ierr = MPI_Comm_size(comm, &size);CHKERRQ(ierr);
  {
     const int          stage=0;
     PetscStageLog      stageLog;
     ierr          = PetscLogGetStageLog(&stageLog);CHKERRQ(ierr);
     eventRegInfo  = stageLog->eventLog->eventInfo;
     eventPerfInfo = stageLog->stageInfo[stage].eventLog->eventInfo;
  }

  /* For each default timer, calculate the (one) nested timer that it corresponds to. */
  maxDefaultTimer =0;
  for (i=0; i<nNestedEvents; i++) {
     size_t nParents                 = nestedEvents[i].nParents;
     PetscLogEvent *dftEvents        = nestedEvents[i].dftEvents;
     int j;
     for (j=0; j<nParents; j++) {
         maxDefaultTimer = MAX(dftEvents[j],maxDefaultTimer);
     }
  }
  ierr = PetscMalloc((maxDefaultTimer+1)*sizeof(NestedEventId),&nstEvents); CHKERRQ(ierr);
  for (dft=0; dft<maxDefaultTimer; dft++) {nstEvents[dft] = 0;}
  for (i=0; i<nNestedEvents; i++) {
     size_t nParents                 = nestedEvents[i].nParents;
     NestedEventId nstEvent          = nestedEvents[i].nstEvent;
     PetscLogEvent *dftEvents        = nestedEvents[i].dftEvents;
     int j;
     for (j=0; j<nParents; j++) { nstEvents[dftEvents[j]] = nstEvent; }
  }

  /* Calculate largest nested event-ID */
  nstMax_local = 0;
  for (i=0; i<nNestedEvents; i++) { if (nestedEvents[i].nstEvent>nstMax_local) {nstMax_local = nestedEvents[i].nstEvent;} }
  ierr = MPI_Allreduce(&nstMax_local, &nstMax, 1, MPI_INT, MPI_MAX, comm);CHKERRQ(ierr);


  /* Initialize all total-times with zero */
  ierr = PetscMalloc((nstMax+1)*sizeof(PetscSelfTimer),&selftimes); CHKERRQ(ierr);
  ierr = PetscMalloc((nstMax+1)*sizeof(PetscSelfTimer),&totaltimes); CHKERRQ(ierr);
  for (nst=0; nst<=nstMax; nst++) { 
      totaltimes[nst].time          = 0;
      totaltimes[nst].flops         = 0;
      totaltimes[nst].numMessages   = 0;
      totaltimes[nst].messageLength = 0;
      totaltimes[nst].numReductions = 0;
      totaltimes[nst].name          = NULL;
  }


  /* Calculate total-times */
  for (i=0; i<nNestedEvents; i++) { 
     const size_t         nParents  = nestedEvents[i].nParents;
     const NestedEventId  nstEvent  = nestedEvents[i].nstEvent;
     const PetscLogEvent *dftEvents = nestedEvents[i].dftEvents;
     int j;
     for (j=0; j<nParents; j++) {
         const PetscLogEvent dftEvent = dftEvents[j];
         totaltimes[nstEvent].time          += eventPerfInfo[dftEvent].time;
         totaltimes[nstEvent].flops         += eventPerfInfo[dftEvent].flops;
         totaltimes[nstEvent].numMessages   += eventPerfInfo[dftEvent].numMessages;
         totaltimes[nstEvent].messageLength += eventPerfInfo[dftEvent].messageLength;
         totaltimes[nstEvent].numReductions += eventPerfInfo[dftEvent].numReductions;
     }
     totaltimes[nstEvent].name    = eventRegInfo[(PetscLogEvent) nstEvent].name;
  }


  /* Initialize: self-times := totaltimes */
  for (nst=0; nst<=nstMax; nst++) { selftimes[nst] = totaltimes[nst]; }

  /* Subtract timed supprocesses from self-times */
  for (i=0; i<nNestedEvents; i++) { 
     const size_t nParents                 = nestedEvents[i].nParents;
     const PetscLogEvent *dftEvents        = nestedEvents[i].dftEvents;
     const NestedEventId *dftParentsSorted = nestedEvents[i].dftParentsSorted;
     int j;
     for (j=0; j<nParents; j++) {
         const PetscLogEvent dftEvent  = dftEvents[j];
         const NestedEventId nstParent = nstEvents[dftParentsSorted[j]];
         selftimes[nstParent].time          -= eventPerfInfo[dftEvent].time;
         selftimes[nstParent].flops         -= eventPerfInfo[dftEvent].flops;
         selftimes[nstParent].numMessages   -= eventPerfInfo[dftEvent].numMessages;
         selftimes[nstParent].messageLength -= eventPerfInfo[dftEvent].messageLength;
         selftimes[nstParent].numReductions -= eventPerfInfo[dftEvent].numReductions;
     }
  }


  /* Clean up */
  ierr = PetscFree(nstEvents); CHKERRQ(ierr);
  ierr = PetscFree(totaltimes); CHKERRQ(ierr);


  /* Set outputs */
  *p_self  = selftimes;
  *p_nstMax = nstMax;

  PetscFunctionReturn(0);
}

/*****************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscPrintSelfTime"
static PetscErrorCode  PetscPrintSelfTime(MPI_Comm comm, FILE *fd, const PetscSelfTimer *selftimes, const unsigned long long nstMax, PetscLogDouble totalTime) 
{
  PetscMPIInt        size,rank;
  PetscErrorCode     ierr;
  unsigned long long i;
  NestedEventId      nst;
  PetscSortItem      *sortSelfTimes;
  PetscLogDouble     *times, *maxTimes;
  PetscEventRegInfo  *eventRegInfo; 
  PetscEventPerfInfo *eventPerfInfo;
  const int          dum_depth = 1, dum_count=1, dum_parentcount=1;
  bool               wasPrinted;

  PetscFunctionBegin;

  ierr = MPI_Comm_rank(comm, &rank);CHKERRQ(ierr);
  ierr = MPI_Comm_size(comm, &size);CHKERRQ(ierr);
  {
     const int          stage=0;
     PetscStageLog      stageLog;
     ierr          = PetscLogGetStageLog(&stageLog);CHKERRQ(ierr);
     eventRegInfo  = stageLog->eventLog->eventInfo;
     eventPerfInfo = stageLog->stageInfo[stage].eventLog->eventInfo;
  }

  ierr = PetscMalloc((nstMax+1)*sizeof(PetscLogDouble),&times); CHKERRQ(ierr);
  ierr = PetscMalloc((nstMax+1)*sizeof(PetscLogDouble),&maxTimes); CHKERRQ(ierr);
  for (nst=0; nst<=nstMax; nst++) { times[nst] = selftimes[nst].time;}
  ierr = MPI_Allreduce(times, maxTimes, nstMax+1, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);
  ierr = PetscFree(times); CHKERRQ(ierr);

  ierr = PetscMalloc((nstMax+1)*sizeof(PetscSortItem),&sortSelfTimes); CHKERRQ(ierr);

  /* Sort the self-timers on basis of the largest time needed */
  for (nst=0; nst<=nstMax; nst++) {
     sortSelfTimes[nst].id  = nst;
     sortSelfTimes[nst].val = maxTimes[nst];
  }
  ierr = PetscFree(maxTimes); CHKERRQ(ierr);
  qsort(sortSelfTimes, nstMax+1, sizeof(PetscSortItem), compareSortItems);

  if (!useXMLFormat) {
    ierr = PetscFPrintf(comm, fd, "\n\nSelf-timers (time is in %% of %0.2f s):\n\n", totalTime); CHKERRQ(ierr);
  } else {
    ierr = PetscXMLStartSection(comm, fd, "selftimertable", "Self-timings"); CHKERRQ(ierr);
    ierr = PetscXMLPutDouble(comm, fd, "totaltime", "", totalTime, "%f"); CHKERRQ(ierr);
  }

  for (i=0; i<=nstMax; i++) {
     if ((sortSelfTimes[i].val/totalTime) >= (threshTime/100.0)) {
        NestedEventId  nstEvent = sortSelfTimes[i].id;
        const char    *name     = eventRegInfo[(PetscLogEvent) nstEvent].name;
        PetscEventPerfInfo selfPerfInfo;

        selfPerfInfo.time          = selftimes[nstEvent].time ;
        selfPerfInfo.flops         = selftimes[nstEvent].flops;
        selfPerfInfo.numMessages   = selftimes[nstEvent].numMessages;
        selfPerfInfo.messageLength = selftimes[nstEvent].messageLength;
        selfPerfInfo.numReductions = selftimes[nstEvent].numReductions;

        ierr = PetscLogPrintNestedLine(comm, fd, selfPerfInfo, dum_count, dum_parentcount, dum_depth, name, totalTime, &wasPrinted); CHKERRQ(ierr); 
        if ((useXMLFormat)&&(wasPrinted)){
          ierr = PetscXMLEndSection(comm, fd, "event"); CHKERRQ(ierr);
        }
     }
  }
  if (useXMLFormat) {
    ierr = PetscXMLEndSection(comm, fd, "selftimertable"); CHKERRQ(ierr);
  }
  ierr = PetscFree(sortSelfTimes); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}


/*****************************************************************/
void PetscLogViewNestedRoutine( PetscViewer *viewer, PetscErrorCode *ierr)
{
  PetscViewer v;
  PetscPatchDefaultViewers_Fortran(viewer,v);
  *ierr = PetscLogViewNested(v);
}

/*****************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscLogViewNested"
PetscErrorCode  PetscLogViewNested(PetscViewer viewer)
{
  FILE               *fd;
  MPI_Comm           comm;
  PetscErrorCode     ierr;
  PetscLogDouble     locTotalTime, globTotalTime;
  PetscMPIInt        rank;
  PetscNestedEventTree *tree;
  PetscSelfTimer     *selftimers;
  unsigned long long  nTimers, nstMax;
  PetscViewerType    vType;
  PetscViewerFormat  vFormat;

  PetscFunctionBegin;
  ierr = PetscViewerGetType(viewer,&vType);CHKERRQ(ierr);
  if (strcmp(vType,PETSCVIEWERASCII) != 0) {
     SETERRQ(PETSC_COMM_WORLD,PETSC_ERR_ARG_OUTOFRANGE,"Viewer must be ASCII viewer");
     SETERRQ2(PETSC_COMM_WORLD, PETSC_ERR_ARG_OUTOFRANGE, 
              "Viewer is of type %s, but should be %s (ASCII)\n", vType, PETSCVIEWERASCII);
  }

  /* Set useXMLFormat that controls the format in all local PetscPrint.. functions */
  ierr = PetscViewerCheckASCII_XMLFormat(viewer, &useXMLFormat);CHKERRQ(ierr);

  ierr = PetscObjectGetComm((PetscObject)viewer,&comm);CHKERRQ(ierr);
  ierr = PetscViewerASCIIGetPointer(viewer,&fd);CHKERRQ(ierr);
  ierr = MPI_Comm_rank(comm, &rank);CHKERRQ(ierr);


  if (!useXMLFormat) { 
    ierr = PetscFPrintf(comm, fd, "\n--------- PETSc Performance Summary: --------------\n\n");CHKERRQ(ierr);
  } else {
    ierr = PetscFPrintf(comm, fd, "<!-- PETSc Performance Summary: -->\n");CHKERRQ(ierr);
    ierr = PetscXMLStartSection(comm, fd, "petscroot", "");CHKERRQ(ierr);
  }
  
  /* Get the total elapsed time, local and global maximum */
  PetscTime(&locTotalTime);  locTotalTime -= petsc_BaseTime;
  ierr = MPI_Allreduce(&locTotalTime, &globTotalTime, 1, MPIU_PETSCLOGDOUBLE, MPI_MAX, comm);CHKERRQ(ierr);

  /* Print global information about this run */
  ierr = PetscPrintExeSpecs(comm, fd);CHKERRQ(ierr);
  ierr = PetscPrintGlobalPerformance(comm, fd, locTotalTime);CHKERRQ(ierr); 
  ierr = PetscPrintPerformanceWarning(comm, fd);CHKERRQ(ierr);  
  
  /* Collect nested timer tree info from all processes */
  ierr = PetscCreateLogTreeNested(comm, fd, &tree, &nTimers);CHKERRQ(ierr);
  ierr = PetscLogNestedPrintTop(comm, fd, tree, nTimers, globTotalTime);CHKERRQ(ierr);
  ierr = PetscLogFreeNestedTree(tree, nTimers);CHKERRQ(ierr);

  /* Calculate self-time for all (not-nested) events */
  ierr = PetscCalcSelfTime(comm, fd, &selftimers, &nstMax);CHKERRQ(ierr);
  ierr = PetscPrintSelfTime(comm, fd, selftimers, nstMax, globTotalTime);CHKERRQ(ierr);
  ierr = PetscFree(selftimers); CHKERRQ(ierr);

  if (!useXMLFormat) { 
    ierr = PetscFPrintf(comm, fd, "\n--------- End of PETSc Performance Summary --------\n\n");CHKERRQ(ierr);
  } else {
    ierr = PetscXMLEndSection(comm, fd, "petscroot");CHKERRQ(ierr);
    ierr = PetscViewerFinalASCII_XML(&viewer);CHKERRQ(ierr);
  }
  PetscViewerDestroy(&viewer);
  PetscFunctionReturn(0);
}

/*****************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscAdminSend"
PetscErrorCode PetscAdminSend(int count, int datatype) {
  PetscErrorCode    ierr;
  PetscFunctionBegin;
  petsc_send_ct++;
  ierr = PetscMPITypeSize(&petsc_send_len,count, MPI_Type_f2c((MPI_Fint) datatype)); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

/*****************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscAdminRecv"
PetscErrorCode PetscAdminRecv(int count, int datatype) {
  PetscErrorCode    ierr;
  PetscFunctionBegin;
  petsc_recv_ct++;
  ierr = PetscMPITypeSize(&petsc_recv_len,count, MPI_Type_f2c((MPI_Fint) datatype)); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

/*****************************************************************/
#undef __FUNCT__
#define __FUNCT__ "PetscAdminReduce"
PetscErrorCode PetscAdminReduce() {
  PetscFunctionBegin;
  petsc_allreduce_ct++;
  PetscFunctionReturn(0);
}


